#
# Full schema for basic all-encompassing workbook checking.
#
# Note the use of the additionalProperties keyword. Because our
# collections (e.g. calculations, scenes, sources) are dicts and
# not arrays, with the key being the index number, to do this schema
# the simplest way, we would have multiple lines in each collection,
# one for each index (1-N), which would be silly.
#
# For the use of "oneOf", this page was very helpful:
#    http://json-schema.org/example2.html
# see also:
#    http://spacetelescope.github.io/understanding-json-schema
#
# This module takes advantage of Python's dictionaries (and sub-dicts)
# to keep things modular, and to create (below) schemas that allow
# validation of a whole workbook, or just of parts (like a source).
#
# Keep in synch with:
#     pandeia/engine/doc/engine_input_api.rst
#
import copy

# Globals
SCHEMA_URL = 'http://json-schema.org/schema#'

# an empty dict ({}) is valid for this schema but not an empty string('')
EMPTY_SCHEMA = {
    'type': 'object',
    '$schema': SCHEMA_URL,
    'additionalProperties': False
}

# This regex matches a string representation of an integer.
INT_REGEX = r'^-?[0-9]+$'
FLOAT_REGEX = r'^-?[0-9]+\.?[0-9]*$'

#-------------------------------------------------------------------------------
# DEFINITIONS OF PARTS - these are not to be used directly during validation,
# but instead incorporated into regular schemas.
#
# These are declared in order of most atomic to most complex
#-------------------------------------------------------------------------------

a_norm_method = {
    'type': 'string',
    'enum': ['integ_infinity', 'surf_scale', 'surf_center'],
}

a_surf_area_units = {
    'oneOf': [
        {'type': 'string', 'enum': ['arcsec^2', 'sr']},
        {'type': 'null'},
    ]
}

a_geom_point = {
    'type': 'object',
    'required': ['geometry'],
    'additionalProperties': False,
    'properties': {
        'geometry': {'type': 'string', 'enum': ['point']},
    },
}

a_geom_power = {
    'type': 'object',
    'required': ['geometry','power_index','r_core','norm_method','surf_area_units'],
    'additionalProperties': False,
    'properties': {
        'geometry':        {'type': 'string', 'enum': ['power']},
        'r_core':          {'type': 'number'},
        'power_index':     {'type': 'number'},
        'norm_method':     a_norm_method,
        'surf_area_units': a_surf_area_units,
    },
}

a_geom_flat_or_g2d = {
    'type': 'object',
    'required': ['geometry','major','minor','norm_method','surf_area_units'],
    'additionalProperties': False,
    'properties': {
        'geometry':        {'type': 'string', 'enum': ['flat','gaussian2d']},
        'major':           {'type': 'number'},
        'minor':           {'type': 'number'},
        'norm_method':     a_norm_method,
        'surf_area_units': a_surf_area_units,
    },
}

a_geom_sersic = { # just like a_geom_flat_or_g2d but with extra sersic_index field
    'type': 'object',
    'required': ['geometry','major','minor','sersic_index','norm_method','surf_area_units'],
    'additionalProperties': False,
    'properties': {
        'geometry':        {'type': 'string', 'enum': ['sersic', 'sersic_scale']},
        'major':           {'type': 'number'},
        'minor':           {'type': 'number'},
        'sersic_index':    {'type': 'number'},
        'norm_method':     a_norm_method,
        'surf_area_units': a_surf_area_units,
    },
}

a_normalization_none = {
    'type': 'object',
    'required': ['type'],
    'additionalProperties': False,
    'properties': {
        'type': {'type': 'string', 'enum': ['none']},
    },
}

a_normalization_at_lambda = {
    'type': 'object',
    'required': ['norm_flux','norm_fluxunit','norm_wave','norm_waveunit','type'],
    'additionalProperties': False,
    'properties': {
        'norm_flux':     {'type': 'number'},
        'norm_fluxunit': {'type': 'string', 'enum': ['flam','fnu','abmag','vegamag','mjy','ujy','njy','jy','MJy']},
        'norm_wave':     {'type': 'number'},
        'norm_waveunit': {'type': 'string', 'enum': ['angstrom','microns']},
        'type':          {'type': 'string', 'enum': ['at_lambda']},
    },
}

a_normalization_instrument = {
    'type': 'object',
    'required': ['bandpass','norm_flux','norm_fluxunit','type'],
    'additionalProperties': False,
    'properties': {
        'bandpass':      {'type': 'string'},
        'norm_flux':     {'type': 'number'},
        'norm_fluxunit': {'type': 'string', 'enum': ['flam','fnu','abmag','vegamag','mjy','ujy','njy','jy','MJy']},
        'type':          {'type': 'string', 'enum': ['hst','jwst','photsys']},
    },
}

an_sed_none = {
    'type': 'object',
    'required': ['sed_type'],
    'additionalProperties': False,
    'properties': {
        'sed_type': {'type': 'string', 'enum': ['no_continuum']},
    },
}

an_sed_flat = {
    'type': 'object',
    'required': ['sed_type','unit'],
    'additionalProperties': False,
    'properties': {
        'sed_type': {'type': 'string', 'enum': ['flat']},
        'unit':     {'type': 'string', 'enum': ['flam','fnu'] },
    },
}

an_sed_blackbody = {
    'type': 'object',
    'required': ['sed_type','temp'],
    'additionalProperties': False,
    'properties': {
        'sed_type': {'type': 'string', 'enum': ['blackbody']},
        'temp':     {'type': 'number'},
    },
}

an_sed_powerlaw = {
    'type': 'object',
    'required': ['index','sed_type','unit'],
    'additionalProperties': False,
    'properties': {
        'index':    {'type': 'number'},
        'sed_type': {'type': 'string', 'enum': ['powerlaw']},
        'unit':     {'type': 'string', 'enum': ['flam','fnu'] },
    },
}

an_sed_keyed = {
    'type': 'object',
    'required': ['key','sed_type'],
    'additionalProperties': False,
    'properties': {
        'key':      {'type': 'string'},
        'sed_type': {'type': 'string'},
    },
}

an_sed_upload = {
    'type': 'object',
    'required': ['sed_type','spectrum_id'],
    'additionalProperties': False,
    'properties': {
        'sed_type':    {'type': 'string', 'enum': ['input']},
        'spectrum_id': {'type': 'number'},
    },
}

a_background_hst = {
    'type': 'object',
    'required': ['airglow','earthshine','zodiacallight'],
    'additionalProperties': False,
    'properties': {
        'airglow': {
            'type': 'object',
            'required': ['choice', 'standard'],
            'additionalProperties': False,
            'properties': {
                'choice':     {'type': 'string', 'enum': ['standard']},
                'standard':   {'type': 'string', 'enum': ['none', 'low', 'average', 'high']}
            }
        },
        'earthshine': {
            'type': 'object',
            'required': ['choice'],
            'additionalProperties': False,
            'properties': {
                'choice':     {'type': 'string', 'enum': ['standard', 'magnitude', 'multiplier']},
                'standard':   {'type': 'string', 'enum': ['none', 'average', 'high', 'extremely_high']},
                'magnitude':  {'type': 'number'},
                'multiplier': {'type': 'number'}
            }
        },
        'zodiacallight':  {
            'type': 'object',
            'required': ['choice'],
            'additionalProperties': False,
            'properties': {
                'choice':     {'type': 'string', 'enum': ['standard', 'magnitude', 'multiplier']},
                'standard':   {'type': 'string', 'enum': ['none', 'low', 'average', 'high']},
                'magnitude':  {'type': 'number'},
                'multiplier': {'type': 'number'}
            }
        }
    }
}

a_background_hst_ir = copy.deepcopy(a_background_hst)
a_background_hst_ir['properties']['helium'] = a_background_hst_ir['properties']['airglow']
del a_background_hst_ir['properties']['airglow']
a_background_hst_ir['properties']['helium']['properties']['standard']['enum'] = ['none', 'average', 'high', 'very_high']
a_background_hst_ir['required'] = ['earthshine','helium','zodiacallight']

a_background_none = {
    'type': 'object',
    'required': ['bg_type'],
    'additionalProperties': False,
    'properties': {
        'bg_type': {'type': 'string', 'enum': ['none']},
    },
}

a_background_dateless = {
    'type': 'object',
    'required': ['bg_type','ra','dec','ra_dec_str'],
    'additionalProperties': False,
    'properties': {
        'bg_type':    {'type': 'string', 'enum': ['low','medium','high']},
        'ra':         {'type': 'number'},
        'dec':        {'type': 'number'},
        'ra_dec_str': {'type': 'string'},
    },
}

a_background_dated = {
    'type': 'object',
    'required': ['bg_type','ra','dec','date'],
    'additionalProperties': False,
    'properties': {
        'bg_type':    {'type': 'string', 'enum': ['low','medium','high','dated']},
        'ra':         {'type': 'number'},
        'dec':        {'type': 'number'},
        'date':       {'type': 'number'},
        'ra_dec_str': {'type': 'string'}, # optional
        'date_str':   {'type': 'string'}, # optional
    },
}

a_background_choice = {
    'type': 'object',
    'oneOf': [a_background_hst, a_background_hst_ir, a_background_none, a_background_dateless, a_background_dated],
}

a_dither_simple = {
    'type': 'array',
    'items': {
        'type': 'object',
        'required': ['x','y'],
        'additionalProperties': False,
        'properties': {
            'x': {'type': 'number'},
            'y': {'type': 'number'},
        }
    }
}

a_dither_w_bg = {
    'type': 'array',
    'items': {
        'type': 'object',
        'required': ['x','y','on_source'],
        'additionalProperties': False,
        'properties': {
            'x':         {'type': 'number'},
            'y':         {'type': 'number'},
            'on_source': {'type': 'array', 'items': {'type': 'boolean'}},
        }
    }
}

a_helper_strategy_background = {
    'type': 'object',
    'required': ["background_subtraction"],

    'properties': {
        'background_subtraction': {'type': 'boolean', },
        'sky_annulus':            {'type': 'array', 'items': {'type': 'number'}},
    },

    # Dep from background_subtraction to sky_annulus
    'oneOf': [
        {
            "properties": {
                'background_subtraction': { 'enum': [True] },
            },
            "required": ["sky_annulus"],
        },
        {
            "properties": {
                'background_subtraction': { 'enum': [False] },
            },
        },
    ],

    # Dep from sky_annulus to background_subtraction
    'dependencies': {
        'sky_annulus': {
            'properties': {
                'background_subtraction': { 'enum': [True] },
            },
        },
    },
}

a_strategy_coronagraphy = {
    'allOf': [
        {
            'type': 'object',
            'required': ['annulus_shape','aperture_size','contrast_azimuth','contrast_separation','method','psf_subtraction','psf_subtraction_source','scene_rotation','target_source','target_type','strat_offset_xy','units'],
            'properties': {
                'annulus_shape':          {'type': 'string', 'enum': ['circular']},
                'aperture_size':          {'type': 'number'},
                'background_subtraction': {'type': 'boolean', 'enum': [True]},
                'contrast_azimuth':       {'type': 'number'},
                'contrast_separation':    {'type': 'number'},
                'method':                 {'type': 'string', 'enum': ['coronagraphy']},
                'psf_subtraction':        {'type': 'string', 'enum': ['no_autoscale','optimal','psf_only','realistic','target_only']},
                'psf_subtraction_source': {'type': 'number'},
                'scene_rotation':         {'type': 'number'},
                'target_source':          {'type': 'string', 'pattern': INT_REGEX},
                'target_type':            {'type': 'string', 'enum': ['source']},
                'strat_offset_xy':        {'type': 'array', 'items': {'type': 'number'}},
                'units':                  {'type': 'string', 'enum': ['arcsec', 'pixel']},
            },
        },
        a_helper_strategy_background,
    ],
}

a_strategy_ifunodinscene = {
    'type': 'object',
    'required': ['aperture_size','dithers','method','reference_wavelength','target_source','target_type','strat_offset_xy','units'],
    'additionalProperties': False,
    'properties': {
        'aperture_size':        {'type': 'number'},
        'dithers':              a_dither_simple,
        'method':               {'type': 'string', 'enum': ['ifunodinscene']},
        'reference_wavelength': {'oneOf': [{'type': 'number'}, {'type': 'null'}]},
        'target_source':        {'type': 'string', 'pattern': INT_REGEX},
        'target_type':          {'type': 'string', 'enum': ['coords','source']},
        'strat_offset_xy':      {'type': 'array', 'items': {'type': 'number'}},
        'units':                {'type': 'string', 'enum': ['arcsec', 'pixel']},
    },
}

a_strategy_ifunodoffscene = {
    'type': 'object',
    'required': ['aperture_size','method','reference_wavelength','target_source','target_type','strat_offset_xy','units'],
    'additionalProperties': False,
    'properties': {
        'aperture_size':        {'type': 'number'},
        'dithers':              a_dither_simple,  # optional for UI, used only by engine
        'method':               {'type': 'string', 'enum': ['ifunodoffscene']},
        'reference_wavelength': {'oneOf': [{'type': 'number'}, {'type': 'null'}]},
        'target_source':        {'type': 'string', 'pattern': INT_REGEX},
        'target_type':          {'type': 'string', 'enum': ['coords','source']},
        'strat_offset_xy':      {'type': 'array', 'items': {'type': 'number'}},
        'units':                {'type': 'string', 'enum': ['arcsec', 'pixel']},
    },
}

a_strategy_imagingapphot = {
    'allOf': [
        {
            'type': 'object',
            'required': ['aperture_size','method','strat_offset_xy','target_source','target_type','units'],
            'properties': {
                'aperture_size':   {'type': 'number'},
                'method':          {'type': 'string', 'enum': ['imagingapphot']},
                'strat_offset_xy': {'type': 'array', 'items': {'type': 'number'}},
                'target_source':   {'type': 'string', 'pattern': INT_REGEX},
                'target_type':     {'type': 'string', 'enum': ['coords','source']},
                'units':           {'type': 'string', 'enum': ['arcsec','pixel']},
            },
        },
        a_helper_strategy_background,
    ],
}

# ifuapphot is just imagingapphot plus the reference_wavelength
a_strategy_ifuapphot = copy.deepcopy(a_strategy_imagingapphot)
a_strategy_ifuapphot['allOf'][0]['required'] = a_strategy_ifuapphot['allOf'][0]['required'] + ['reference_wavelength',]
a_strategy_ifuapphot['allOf'][0]['properties']['method']['enum'] = ['ifuapphot']
a_strategy_ifuapphot['allOf'][0]['properties']['reference_wavelength'] = {'oneOf': [{'type': 'number'}, {'type': 'null'}]}

a_strategy_imagingscanapphot = {
    'allOf': [
        {
            'type': 'object',
            'required': ['method','scan_extraction_length','scan_extraction_width','strat_offset_xy','target_source','target_type','units'],
            'properties': {
                'method':                 {'type': 'string', 'enum': ['imagingscanapphot']},
                'scan_extraction_length': {'type': 'string', 'enum': ['pixel','scanlength']},
                'scan_extraction_width':  {'type': 'number'},
                'strat_offset_xy':        {'type': 'array', 'items': {'type': 'number'}},
                'target_source':          {'type': 'string', 'pattern': INT_REGEX},
                'target_type':            {'type': 'string', 'enum': ['coords','source']},
                'units':                  {'type': 'string', 'enum': ['arcsec','pixel']},
            },
        },
        a_helper_strategy_background,
    ],
}

a_strategy_msaapphot = {
    'allOf': [
        {
            'type': 'object',
            'required': ['aperture_size','dithers','method','strat_offset_xy','target_source','target_type','units'],
            'properties': {
                'aperture_size':   {'type': 'number'},
                'dithers':         a_dither_simple,
                'method':          {'type': 'string', 'enum': ['msaapphot']},
                'strat_offset_xy': {'type': 'array', 'items': {'type': 'number'}},
                'target_source':   {'type': 'string', 'pattern': INT_REGEX},
                'target_type':     {'type': 'string', 'enum': ['source']},
                'units':           {'type': 'string', 'enum': ['arcsec', 'pixel']},
            },
        },
        a_helper_strategy_background,
    ],
}

a_strategy_msafullapphot = {
    'type': 'object',
    'required': ['background_subtraction','dithers','method','reference_wavelength','target_source','target_type','units','strat_offset_xy'],
    'additionalProperties': False,
    'properties': {
        'background_subtraction':   {'type': 'boolean', },
        'dithers':              a_dither_w_bg,
        'method':               {'type': 'string', 'enum': ['msafullapphot']},
        'reference_wavelength': {'oneOf': [{'type': 'number'}, {'type': 'null'}]},
        'target_source':        {'type': 'string', 'pattern': INT_REGEX},
        'target_type':          {'type': 'string', 'enum': ['source']},
        'strat_offset_xy':      {'type': 'array', 'items': {'type': 'number'}},
        'units':                {'type': 'string', 'enum': ['arcsec', 'pixel']},
    },
}

# msashutterapphot is just msafullapphot without the reference_wavelength
a_strategy_msashutterapphot = copy.deepcopy(a_strategy_msafullapphot)
a_strategy_msashutterapphot['required'] = [rr for rr in a_strategy_msafullapphot['required'] if rr != 'reference_wavelength']
del a_strategy_msashutterapphot['properties']['reference_wavelength']
a_strategy_msashutterapphot['properties']['method']['enum'] = ['msashutterapphot']

a_strategy_soss = {
    'type': 'object',
    'required': ['background_subtraction','method','order','reference_wavelength'], # aperture_size, units
    'additionalProperties': False,
    'properties': {
        'background_subtraction': {'type': 'boolean', 'enum': [False]},
        'method':                 {'type': 'string', 'enum': ['soss']},
        'order':                  {'type': 'number'},
        'reference_wavelength':   {'type': 'number'},
    },
}

a_strategy_specapphot = {
    'allOf': [
        {
            'type': 'object',
            'required': ['aperture_size','method','reference_wavelength','target_source','target_type','strat_offset_xy','units'],
            'properties': {
                'aperture_size':        {'type': 'number'},
                'method':               {'type': 'string', 'enum': ['specapphot']},
                'reference_wavelength': {'oneOf': [{'type': 'number'}, {'type': 'null'}]},
                'target_source':        {'type': 'string', 'pattern': INT_REGEX},
                'target_type':          {'type': 'string', 'enum': ['coords','source']},
                'strat_offset_xy':      {'type': 'array', 'items': {'type': 'number'}},
                'units':                {'type': 'string', 'enum': ['arcsec','pixel']},
            },
        },
        a_helper_strategy_background,
    ],
}

# specsegmentedphot is just specapphot without aperture_size
a_strategy_specsegmentedphot = copy.deepcopy(a_strategy_specapphot)
a_strategy_specsegmentedphot['allOf'][0]['required'].remove('aperture_size')
del a_strategy_specsegmentedphot['allOf'][0]['properties']['aperture_size']
a_strategy_specsegmentedphot['allOf'][0]['properties']['method']['enum'] = ['specsegmentedphot']

# taspec is just specsegmentedphot without reference_wavelength but with acq_mode and stripe (those optional since fuv does not use)
a_strategy_taspec = copy.deepcopy(a_strategy_specsegmentedphot)
a_strategy_taspec['allOf'][0]['required'].remove('reference_wavelength')
del a_strategy_taspec['allOf'][0]['properties']['reference_wavelength']
a_strategy_taspec['allOf'][0]['properties']['method']['enum'] = ['taspec']
a_strategy_taspec['allOf'][0]['properties']['acq_mode'] = {'type': 'string', 'enum': ['acq_search','acq_peakd','acq_peakxd']}
a_strategy_taspec['allOf'][0]['properties']['stripe'] = {'type': 'string', 'enum': ['a', 'b', 'c']}

a_strategy_specscanapphot = {
    'allOf': [
        {
            'type': 'object',
            'required': ['method','reference_wavelength','scan_extraction_length','target_source','target_type','strat_offset_xy','units'],
            'properties': {
                'method':                 {'type': 'string', 'enum': ['specscanapphot']},
                'reference_wavelength':   {'oneOf': [{'type': 'number'}, {'type': 'null'}]},
                'scan_extraction_length': {'type': 'string', 'enum': ['pixel','scanlength']},
                'target_source':          {'type': 'string', 'pattern': INT_REGEX},
                'target_type':            {'type': 'string', 'enum': ['coords','source']},
                'strat_offset_xy':        {'type': 'array', 'items': {'type': 'number'}},
                'units':                  {'type': 'string', 'enum': ['arcsec','pixel']},
            },
        },
        a_helper_strategy_background,
    ],
}

a_strategy_taphot = {
    'type': 'object',
    'required': ['background_subtraction','method','target_source','target_type','strat_offset_xy','units'],
    'additionalProperties': False,
    'properties': {
        'background_subtraction': {'type': 'boolean'},
        'method':                 {'type': 'string', 'enum': ['taphot']},
        'target_source':          {'type': 'string', 'pattern': INT_REGEX},
        'target_type':            {'type': 'string', 'enum': ['coords','source']},
        'strat_offset_xy':        {'type': 'array', 'items': {'type': 'number'}},
        'units':                  {'type': 'string', 'enum': ['arcsec','pixel']},
    },
}

# tarectphot is just taphot plus acq_mode
a_strategy_tarectphot = copy.deepcopy(a_strategy_taphot)
a_strategy_tarectphot['required'].append('acq_mode')
a_strategy_tarectphot['properties']['acq_mode'] = {'type': 'string', 'enum': ['acq_image','acq_search']}
a_strategy_tarectphot['properties']['method']['enum'] = ['tarectphot']

# taboxphot is just tarectphot plus acq_type and aperture_size
a_strategy_taboxphot = copy.deepcopy(a_strategy_tarectphot)
a_strategy_taboxphot['required'].append('acq_type')
a_strategy_taboxphot['required'].append('aperture_size')
a_strategy_taboxphot['properties']['acq_mode'] = {'type': 'string', 'enum': ['acq_image','acq_peak']}
a_strategy_taboxphot['properties']['acq_type'] = {'type': 'string', 'enum': ['point','extended']}
a_strategy_taboxphot['properties']['aperture_size'] = {'type': 'number'}
a_strategy_taboxphot['properties']['method']['enum'] = ['taboxphot']


a_strategy_choice = {
    'type': 'object',
    'oneOf': [a_strategy_coronagraphy, a_strategy_ifuapphot, a_strategy_ifunodinscene, a_strategy_ifunodoffscene, a_strategy_imagingapphot,
              a_strategy_imagingscanapphot, a_strategy_msaapphot, a_strategy_msafullapphot, a_strategy_msashutterapphot,
              a_strategy_soss, a_strategy_specapphot, a_strategy_specscanapphot, a_strategy_specsegmentedphot, a_strategy_taboxphot,
              a_strategy_taphot, a_strategy_tarectphot, a_strategy_taspec],
}

a_scene_element = {
    'type': 'object',
    'required': ['deleted','name','orientation','source_arg_blob','source_id','x_offset','y_offset'],
    'additionalProperties': False,
    'properties': {
        'deleted':         {'type': 'number'},
        'name':            {'type': 'string'},
        'orientation':     {'type': 'number'},
        'source_arg_blob': EMPTY_SCHEMA, # unused at current
        'source_id':       {'type': 'number'},
        'x_offset':        {'type': 'number'},
        'y_offset':        {'type': 'number'},
    },
}

a_scene = {
    'type': 'object',
    'required': ['characteristics','deleted','desc','name','scene_elements'],
    'additionalProperties': False,
    'properties': {
        'characteristics': EMPTY_SCHEMA, # unused at current
        'deleted':         {'type': 'number'},
        'desc':            {'type': 'string'},
        'name':            {'type': 'string'},
        'scene_elements': {
            'type': 'object',
            'additionalProperties': a_scene_element,
            'properties': {},
        },
    },
}

a_source_line = {
    'type': 'object',
    'required': ['center','emission_or_absorption','name','profile','strength','width'],
    'additionalProperties': False,
    'properties': {
        'center':                 {'type': 'number'},
        'emission_or_absorption': {'type': 'string', 'enum': ['emission','absorption']},
        'name':                   {'type': 'string'},
        'profile':                {'type': 'string', 'enum': ['gaussian']},
        'strength':               {'type': 'number'},
        'width':                  {'type': 'number'},
    },
}

a_source_characteristics = {
    'type': 'object',
    'required': ['shape','spectrum'],
    'additionalProperties': False,
    'properties': {
        'shape': {
            'type': 'object',
            'oneOf': [a_geom_point, a_geom_flat_or_g2d, a_geom_sersic, a_geom_power],
        },
        'spectrum': {
            'type': 'object',
            'required': ['lines','normalization','redshift','sed','extinction'],
            'additionalProperties': False,
            'properties': {
                'extinction_first': {'type': 'boolean'},
                'lines': {'type': 'array', 'items': a_source_line},
                'normalization': {
                    'type': 'object',
                    'oneOf': [a_normalization_none, a_normalization_at_lambda, a_normalization_instrument],
                },
                'redshift': {'type': 'number'},
                'sed': {
                    'type': 'object',
                    'oneOf': [an_sed_none, an_sed_blackbody, an_sed_flat, an_sed_powerlaw, an_sed_keyed, an_sed_upload],
                },
                'extinction': {
                    'law':      {'type': 'string'},
                    'bandpass': {'type': 'string'},
                    'unit':     {'type': 'string'},
                    'value':    {'type': 'float'}
                }
            }
        }
    }
}

a_source = {
    'type': 'object',
    'required': ['characteristics','deleted','name'],
    'additionalProperties': False,
    'properties': {
        'characteristics': a_source_characteristics,
        'deleted': {'type': 'number'},
        'name':    {'type': 'string'},
    },
}

a_detector_hst = {
    'type': 'object',
    'required': ['calculate_snr','snr','time'],
    'additionalProperties': False,
    'properties': {
        'bin_dispersion':  {'type': 'number'}, # optional
        'bin_spatial':     {'type': 'number'}, # optional
        'calculate_snr':   {'type': 'boolean'},
        'dark_level':      {'type': 'string'}, # optional
        'gain':            {'type': 'string', 'pattern': FLOAT_REGEX}, # optional
        'fuv_glow_region': {'type': 'string'}, # optional
        'nexp':            {'type': 'number'}, # optional
        'nsplit':          {'type': 'number'}, # optional
        'postflash':       {'type': 'number'}, # optional
        'scan_choice':     {'type': 'string', 'enum': ['rate', 'time']}, # optional
        'scan_length':     {'type': 'number'}, # optional
        'snr':             {'type': 'number'},
        'time':            {'type': 'number'},
    },
}

a_detector_jwst = {
    'type': 'object',
    'required': ['nexp','ngroup','nint','readout_pattern'],
    'additionalProperties': False,
    'properties': {
        'nexp':            {'type': 'number'},
        'ngroup':          {'type': 'number'},
        'nint':            {'type': 'number'},
        'readout_pattern': {'type': 'string'},
        'subarray':        {'type': 'string'}, # optional
    },
}

a_detector_for_st = {
    'type': 'object',
    'oneOf': [a_detector_hst, a_detector_jwst],
}

a_camera_config = {
    'type': 'object',
    'required': ['detector','instrument'],
    'additionalProperties': False,
    'properties': {
        'detector': a_detector_for_st,
        'instrument': {
            'type': 'object',
            'required': ['aperture','disperser','filter'],
            'additionalProperties': False,
            'properties': {
                'aperture':         {'type': 'string'},
                'cenwave':          {'type': 'string'}, # optional, HST only, may break all this up
                'slit':             {'type': 'string'}, # optional, HST only, may break all this up
                'detector':         {'type': 'string'}, # optional, both HST and JWST use
                'disperser':        {'oneOf': [{'type': 'string'}, {'type': 'null'}]},
                'filter':           {'oneOf': [{'type': 'string'}, {'type': 'null'}]},
                'paired_filter':    {'oneOf': [{'type': 'string'}, {'type': 'null'}]}, # optional, JWST only
                'shutter_location': {'type': 'string'}, # optional
                'slitlet_shape': {'type': 'string'} # optional
            }
        }
    }
}

a_client_data_other_strategies_dict = {
    'type': 'object',
    'required': [],
    'additionalProperties': False,
    'properties': { # need list only those used so far for calcs with multiple possible strategies
        'ifuapphot': a_strategy_ifuapphot,
        'ifunodinscene': a_strategy_ifunodinscene,
        'ifunodoffscene': a_strategy_ifunodoffscene,
        'msafullapphot': a_strategy_msafullapphot,
        'specapphot': a_strategy_specapphot,
    }
#   Prefer to do it this way, to not hard-code the list of strats here, but can't get it to work
#   'properties': { '^[a-z][a-z]*': a_strategy_choice }
}

a_client_data = {
    'oneOf': [
        {'type': 'null'},
        {
            'type': 'object',
            'required': [], # everything in client_data is optional
            'additionalProperties': False,
            'properties': {
                'cause': {'type': 'string'},
                'note': {'type': 'string'},
                'last_ra_dec_str': {'type': 'string'},
                'other_strategies': a_client_data_other_strategies_dict,
                'reserved': {'type': 'string'}, # have in place in case a quick patch is needed
                'sent': {'type': 'number'},
            }
        },
    ]
}

an_instrument_name_hst =  {'type': 'string', 'enum': ['acs','cos','stis','wfc3']}
an_instrument_name_jwst = {'type': 'string', 'enum': ['miri','nircam','niriss','nirspec']}

an_instrument_name = {
    'oneOf': [an_instrument_name_hst, an_instrument_name_jwst],
}

a_calculation = {
    'type': 'object',
    'required': ['apt','background','camera_config','deleted','insmode','instrument','name','scene_id','strategy','strategy_args'],
    'additionalProperties': False,
    'properties': {
        'apt':             {'type': 'string'},
        'background':      a_background_choice,
        'camera_config':   a_camera_config,
        'client_data':     a_client_data, # optional
        'deleted':         {'type': 'number'},
        'insmode':         {'type': 'string'},
        'instrument':      an_instrument_name,
        'name':            {'type': 'string'},
        'scene_id':        {'type': 'number'},
        'strategy':        {'type': 'string'},
        'strategy_args':   a_strategy_choice,
    }
}


#-------------------------------------------------------------------------------
# Complete, actionable schemas to be used for validation
#-------------------------------------------------------------------------------

WORKBOOK_SCHEMA = {
    '$schema': SCHEMA_URL,
    'type': 'object',
    'required': ['calculations','created','deleted','desc','hash','id','name','note',
                 'proposal_id','proposal_state','scenes','sources','test_mode'],
    'additionalProperties': False,
    'properties': {
        'created':        {'type': 'number'},
        'default_scene':  {'type': 'number'},
        'deleted':        {'type': 'number'},
        'desc':           {'type': 'string'},
        'hash':           {'type': 'string'},
        'id':             {'type': 'number'},
        'name':           {'type': 'string'},
        'note':           {'type': 'string'},
        'proposal_id':    {'type': 'string'},
        'proposal_state': {'type': 'string'},
        'test_mode':      {'type': 'string'},
        'calculations': {
            'type': 'object',
            'additionalProperties': {'$ref': '#/definitions/a_calculation'},
            'properties': {},
        },
        'scenes': {
            'type': 'object',
            'additionalProperties': a_scene,
            'properties': {},
        },
        'sources': {
            'type': 'object',
            'additionalProperties': a_source,
            'properties': {},
        },
    },

    # definitions: anything listed here can be used in the $ref syntax
    'definitions': {

        # definitions from dicts defined above, in order of increasing complexity
        'a_norm_method':                a_norm_method,
        'a_surf_area_units':            a_surf_area_units,
        'a_geom_point':                 a_geom_point,
        'a_geom_flat_or_g2d':           a_geom_flat_or_g2d,
        'a_geom_sersic':                a_geom_sersic,
        'a_geom_power':                 a_geom_power,
        'a_normalization_none':         a_normalization_none,
        'a_normalization_at_lambda':    a_normalization_at_lambda,
        'a_normalization_instrument':   a_normalization_instrument,
        'an_sed_none':                  an_sed_none,
        'an_sed_flat':                  an_sed_flat,
        'an_sed_blackbody':             an_sed_blackbody,
        'an_sed_powerlaw':              an_sed_powerlaw,
        'an_sed_keyed':                 an_sed_keyed,
        'a_background_hst':             a_background_hst,
        'a_background_hst_ir':          a_background_hst_ir,
        'a_background_none':            a_background_none,
        'a_background_dateless':        a_background_dateless,
        'a_background_dated':           a_background_dated,
        'a_background_choice':          a_background_choice,

        # strategies
        'a_helper_strategy_background': a_helper_strategy_background,
        'a_strategy_coronagraphy':      a_strategy_coronagraphy,
        'a_strategy_ifuapphot':         a_strategy_ifuapphot,
        'a_strategy_ifunodinscene':     a_strategy_ifunodinscene,
        'a_strategy_ifunodoffscene':    a_strategy_ifunodoffscene,
        'a_strategy_imagingapphot':     a_strategy_imagingapphot,
        'a_strategy_imagingscanapphot': a_strategy_imagingscanapphot,
        'a_strategy_msaapphot':         a_strategy_msaapphot,
        'a_strategy_msafullapphot':     a_strategy_msafullapphot,
        'a_strategy_msashutterapphot':  a_strategy_msashutterapphot,
        'a_strategy_soss':              a_strategy_soss,
        'a_strategy_specapphot':        a_strategy_specapphot,
        'a_strategy_specscanapphot':    a_strategy_specscanapphot,
        'a_strategy_specsegmentedphot': a_strategy_specsegmentedphot,
        'a_strategy_taboxphot':         a_strategy_taboxphot,
        'a_strategy_taphot':            a_strategy_taphot,
        'a_strategy_tarectphot':        a_strategy_tarectphot,
        'a_strategy_taspec':            a_strategy_taspec,
        'a_strategy_choice':            a_strategy_choice,

        # scenes/sources
        'a_scene_element':              a_scene_element,
        'a_scene':                      a_scene,
        'a_source':                     a_source,

        # config
        'a_detector_for_st':            a_detector_for_st,
        'a_detector_hst':               a_detector_hst,
        'a_detector_jwst':              a_detector_jwst,
        'a_camera_config':              a_camera_config,
        'an_instrument_name':           an_instrument_name,
        'an_instrument_name_hst':       an_instrument_name_hst,
        'an_instrument_name_jwst':      an_instrument_name_jwst,

        # calculations
        'a_calculation':                a_calculation,
    },
}

# These are the constant schema objects intended for general use.
# For the following objects, use our dict but add schema entry (shallow copy is good enough)

SOURCE_CHAR_SCHEMA = a_source_characteristics.copy()
SOURCE_CHAR_SCHEMA['$schema'] = SCHEMA_URL

CALC_BG_SCHEMA = a_background_choice.copy()
CALC_BG_SCHEMA['$schema'] = SCHEMA_URL

CALC_STRATEGY_SCHEMA = a_strategy_choice.copy()
CALC_STRATEGY_SCHEMA['$schema'] = SCHEMA_URL

CALC_CAMERA_SCHEMA = a_camera_config.copy()
CALC_CAMERA_SCHEMA['$schema'] = SCHEMA_URL

CALC_CLIENT_DATA_SCHEMA = a_client_data.copy()
CALC_CLIENT_DATA_SCHEMA['$schema'] = SCHEMA_URL



""" WFC3IR-specific functions

"""



from .. import scan_mode

from ....engine.imager import Imager
from ....engine.spectrograph import Spectrograph
from ....engine.string_constants import RECTANGLE, SNR_REGION, PIX_REGION
from ....engine.area_over_which import SpecPixAOW


def make_snr_region(instrument, *args, **kwargs):
    """
    WFC3IR in scan mode uses specialized SNR and
    1-pix regions.

    """
    rgn = None

    if isinstance(instrument, Imager) and instrument.scan_mode:

        rgn = scan_mode.make_rectangular_img_snr_region(args, instrument)

    elif isinstance(instrument, Spectrograph) and instrument.scan_mode:

        height = args[0]
        units = args[1]
        obswave = args[2]

        if obswave is None:
            raise ValueError("Must specify obswave.")

        rgn = SpecPixAOW(SNR_REGION, height, units, obswave,
                         instrument._pixels_per_resel,
                         instrument.detector.pixel_angheight,
                         instrument.detector.pixel_angwidth)
        pix_rgn = SpecPixAOW(PIX_REGION, height, units, obswave,
                             1,
                             instrument.detector.pixel_angheight,
                             instrument.detector.pixel_angwidth)

        instrument.areas_of_interest[PIX_REGION] = pix_rgn

    return rgn


def get_configuration_data(ei):
    """ Load instrument and detector data from config files.

        This overrides the base function in order to handle
        WFC3IR specifics.

    Parameters
    ----------
    ei: dict
      engine input

    Returns
    -------
    iconfig: dict
      dictionary of instrument data
    dconfig: dict
      dictionary of detector data
    custom_config: dict
      dictionary with custom data

    """
    return scan_mode.get_scan_configuration_data(ei)


def modify_configuration(instrument, custom_config):
    """ The scan_mode flag must be propagated into the instrument
        instance, and also from the raw table data into instances
        of EE table.

    Parameters
    ----------
    instrument: Imager or Spectrograph
      Instance being configured
    custom_config: dictionary
      custom configuration info

    """
    scan_mode.modify_scan_configuration(instrument, custom_config)



#wfc3ir  limits

# This file is read when the instruments/data.py file is executed
# as part of the instrument module import. The resulting dictionary
# is used to initialize instances of Limit that will be used to
# check any subsequent WFC3IR calculation.

# allsub defines what the magic value "all" should expand into
allsub = {'mode': ['imaging', 'spectroscopic', 'scimaging', 'scspectroscopic'],
          'detector': ['ir']
}

# limit_spec contains the actual limit definitions. It is a
# dictionary {limit shortname : {definition of limit} }.
# See user's guide for detailed requirements of the limit definition.

# Message strings may use the following variable names for substitution:
#  varname = value of the variable specified in "varname"
#  threshold = value specified in "threshold"
#  percentval = percentage expression of (1-subthresh)
#

limit_spec = {

    # Saturation values (brightpix) are taken from the WFC3 Instrument
    # Handbook for Cycle 19, section 7.9.1.
    # The value quoted there is ~78,000e, which is an average over the
    # detector. The more conservative value of 70,000e used in the ETC
    # is a lower-limit to the spatially-variable saturation level.
    'brightpix': {
        'varname': 'brightest_pixel_counts',
        'mode': 'all',
        'message': 'This observation has total electrons per pixel %(varname)g, which exceeds the saturation limit %(threshold)g. Persistence may be an issue with this observation (see handbook for details).'
        ,
        'detector': 'all',
        'threshold': 70000,
        'subthresh': 0.7,
        'submessage': 'This exposure (%(varname)g) exceeds %(percentval)d percent of the detector saturation limit (%(threshold)g)'
    }, #end brightpix

    # mintime threshold value reference: HST Phase II Proposal Instructions
    # for Cycle 18, section 14.3.6, table 14.4 (the entry for sample 1 with
    # a 64x64 image size).
    'mintime' : {
        'varname' : 'time',
        'message' : 
                    '''This calculation for %(varname)g sec uses less than the minimum 
                       exposure time (%(threshold)g sec) for this detector.  See the 
                       Phase II Proposal Instructions for the timing sequences of subarray 
                       and full frame WFC3/IR exposures.''',
        'mincheck' : True,
        'mode' : 'all',
        'detector' : 'all',
        'per_read' : True,
        'threshold': 2.932,
       }, # end mintime

    'maxtime' : {
        'varname' : 'time',
        'message' : '''Result of the calculation %(varname)g is more than the maximum
                    exposure time for this detector (%(threshold)g) ''',
        'mode' : 'all',
        'detector' : 'all',
        'per_read' : True,
        'threshold': 3600.,
       }, #end maxtime

    'maxleak' : {
        'varname' : 'total_leak',
        'message' : '''This observation has a significant filter leak where
                       %(varname).3g percent of the signal comes from outside the prescribed
                       filter bandpass.''',
        'mode' : ['imaging', 'scimaging'],
        'detector' : 'all',
        'per_read' : False,
        'threshold' : 5.0,
        }, # end maxleak

    'zerotargetflux' : {
        'varname' : 'target_total_rate',
        'mincheck' : True,
        'mode' : ['imaging', 'scimaging'],
        'detector' : 'all',
        'per_read' : False,
        'threshold' : 1.E-25,
        'varieties': (
            {'mode'   : 'imaging',
             'message': '''Zero target flux in the observation bandpass''',},
            {'mode'   : 'spectroscopic',
             'message': '''Zero target flux at the observation wavelength''',},
        ),
       }, # end zerotargetflux

    'maxscanrate' : {
        'mode' : ['scimaging', 'scspectroscopic'],
        'detector' : 'all',
        'varname' : 'scan_rate',
        'threshold' : 7.84,
        'message' : '''Scan rate must be lower than %(threshold).3g arcsec/sec''',
        }, # end maxscanrate

    'gyroscanrate' : {
        'mode' : ['scimaging', 'scspectroscopic'],
        'detector' : 'all',
        'varname' : 'scan_rate',
        'threshold' : 5.,
        'message' : '''   'Scan rates larger than %(threshold).3g arcsec/sec are limited to gyro mode only''',
        }, # end gyroscanrate

} #end limit_spec



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import sys
import numpy as np

from astropy import units as u
import synphot as syn
from synphot.models import Empirical1D, ConstFlux1D
import stsynphot as stsyn
from astropy.units import Quantity
from astropy.modeling.models import BlackBody as Black_Body

from .config import DefaultConfig
from .custom_exceptions import DataConfigurationError, EngineInputError, SynphotError, DataError
from .pandeia_warnings import sed_warning_messages as warning_messages
from .utils import recursive_subclasses, merge_data
from .constants import PANDEIA_WAVEUNITS, PANDEIA_FLUXUNITS
from . import config as cf
from pandeia.engine.io_utils import read_json

default_refdata_directory = cf.default_refdata_directory

class SED(DefaultConfig):

    """
    Base class for configuration of spectral energy distributions
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        """
        How an SED is configured depends on the defaults and any input configuration.

        Parameters
        ----------
        webapp: bool
            Determines whether strict API checks should be performed
        config: dict
            Extra configuration information in engine input API dict format
        **kwargs: keyword/value pairs
            Extra configuration information in kwargs format
        """
        self.webapp = webapp

        # API checks are handled below so don't do it here...
        DefaultConfig.__init__(self, webapp=False, config=config, **kwargs)

        if not hasattr(self, "sed_type"):
            self.sed_type = self.__class__.__name__.lower()

        sed_config = {}
        # the SED class requires some settings (particularly for the analytic types) that
        # are only stored in refdata. Load them here.
        if self.sed_type in self.sed_types:
            sed_config = self.sed_types[self.sed_type]
            sed_config = self._load_dict(sed_config,"required", msg="Missing required SED data")
            sed_config = self._load_dict(sed_config,"api", msg="Missing required SED API definitions")
            if not self.webapp:
                if "defaults" in sed_config:
                    type_defaults = sed_config.pop("defaults")
                    sed_config = merge_data(sed_config, type_defaults)
                else:
                    msg = "No defaults defined for SED format %s" % self.sed_type
                    raise DataConfigurationError(value=msg)

        self.__dict__ = merge_data(sed_config, self.__dict__)

        # do some sanity checks to make sure inputs make sense
        try:
            self._sanity_checks()
        except AttributeError as e:
            self.warnings['no_sanity_check'] = warning_messages['no_sanity_check'] % (self.__class__.__name__, e)

        # we need to run the API checks here after merging in the SED type-specific defaults
        if self.webapp:
            all_config = merge_data(config, dict(**kwargs))
            self._api_checks(all_config)

    def _get_config(self):
        """
        Read default configuration from JSON

        Returns
        -------
        config: dict
            All desired class attributes should have defaults defined in the config file
        """
        # use this trick to key the configuration file name off the name of the instantiated subclass
        ref_dir = os.path.join(default_refdata_directory, "sed")
        config = read_json(os.path.join(ref_dir, "config.json"), raise_except=True)
        # JETC-3099: Do not load defaults if webapp==True
        if self.webapp:
            config.pop("defaults", None)
        else:
            config = self._load_dict(config, "defaults")

        # nuke some unused things that are there for UI use only
        for k in ["sed_groupings"]:
            del config[k]
        return config


class Input(SED):

    """
    Base class for SEDs provided as arrays
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        """
        Parameters
        ----------
        webapp: bool
            Determines whether strict API checks should be performed
        config: dict
            Extra configuration information in engine input API dict format
        **kwargs: keyword/value pairs
            Extra configuration information in kwargs format
        """
        SED.__init__(self, webapp=webapp, config=config, **kwargs)
        self.wave, self.flux = self.spectrum
        if isinstance(self.wave, (list, tuple)):
            self.wave = np.array(self.wave)
        if isinstance(self.flux, (list, tuple)):
            self.flux = np.array(self.flux)
        self.wave = self.wave << PANDEIA_WAVEUNITS
        self.flux = self.flux << PANDEIA_FLUXUNITS

        if np.any(np.diff(self.wave) < 0):
            indices = np.where(np.diff(self.wave) < 0)[0]
            msg = "Input spectrum must be sorted by wavelength. Out-of-order indices: %s" % repr(indices)
            raise EngineInputError(value=msg)

        if np.any(np.diff(self.wave) == 0.0):
            indices = np.where(np.diff(self.wave) == 0.0)[0]
            msg = "Input spectrum contains duplicate wavelengths. Duplicate indices: %s" % repr(indices)
            raise EngineInputError(value=msg)

        self.wmin = self.wave.min()
        self.wmax = self.wave.max()

    def _sanity_checks(self):
        """
        Make sure we're given a spectrum and it has elements for wavelength and flux.
        We don't do any unit handling (yet) so units are assumed to be PANDEIA_WAVEUNITS
        and PANDEIA_FLUXUNITS
        """
        if not hasattr(self, "spectrum"):
            msg = "Missing input spectrum."
            raise DataConfigurationError(value=msg)

        if len(self.spectrum) != 2:
            msg = "Must provide input spectrum as a 2 element list-like where 0th index is wavelength and 1st is flux."
            raise EngineInputError(value=msg)


class Analytic(SED):

    """
    Base class for analytic SEDs
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        """
        Parameters
        ----------
        webapp: bool
            Determines whether strict API checks should be performed
        z: float
            Used to adjust wmin/wmax to account for source redshift.
        config: dict
            Extra configuration information in engine input API dict format
        **kwargs: keyword/value pairs
            Extra configuration information in kwargs format
        """
        SED.__init__(self, webapp=webapp, config=config, **kwargs)
        # the wmin and wmax configuration values are in the JWST frame. we need to back out
        # the redshift to calculate the analytic spectra in the correct rest frame.
        self.wmin /= 1.0 + self.z
        self.wmax /= 1.0 + self.z
        self.wave = self.create_wave()
        self.flux = self.create_flux()

    def _sanity_checks(self):
        """
        Check various configuration items to make sure they make sense and raise exception if not.
        """
        if hasattr(self, "unit") and hasattr(self, "units"):
            if self.unit not in self.units:
                msg = "Unsupported unit, %s, for analytic spectrum type %s" % (self.unit, self.__class__.__name__)
                raise EngineInputError(value=msg)

        for par in ["wmin", "wmax", "sampling", "z"]:
            if not hasattr(self, par):
                msg = "%s spectrum missing %s." % (self.__class__.__name__, par)
                raise DataConfigurationError(value=msg)

    def create_wave(self):
        # Create a wavelength grid with constant lambda/delta_lambda
        try:
            k = -self.sampling * np.log(self.wmin)
            nw = int(self.sampling * np.log(self.wmax) + k)
            # nw + 1 almost spans the wavelength range, while nw + 2 guarantees 
            # the highest wavelength will be above the requested range and
            # therefore never require any extrapolation.
            wave = np.exp((np.arange(nw + 2) - k) / self.sampling)
        except ValueError as e:
            msg = "Error creating wavelength grid for wmin=%f, wmax=%s with sampling=%d: %s" % (
                self.wmin,
                self.wmax,
                self.sampling,
                e
            )
            raise EngineInputError(value=msg)
        return wave << PANDEIA_WAVEUNITS

    def create_flux(self):
        flux = np.zeros(len(self.wave))
        return flux << PANDEIA_FLUXUNITS


class Powerlaw(Analytic):

    """
    Class implementing a power-law continuum of the form Flux ~ lam ** index
    """

    def _sanity_checks(self):
        """
        Run sanity checks specify to power-law SEDs
        """
        for par in ['unit', 'units', 'index']:
            if not hasattr(self, par):
                msg = "Missing required parameter %s for %s." % (par, self.__class__.__name__)
                raise DataConfigurationError(value=msg)

    def create_flux(self):
        # effectively normalize to 1 mJy at 1 micron and use Normalize() to scale from there
        flux = self.wave.value ** self.index
        if self.unit == 'flam':
            try:
                # This will be normalized anyway, so the change in units won't matter
                sp = syn.spectrum.SourceSpectrum(Empirical1D, points=self.wave << PANDEIA_WAVEUNITS, lookup_table=flux << syn.units.FLAM)
                flux = sp(sp.waveset).value
            except Exception as e:
                msg = "Error converting flam units to Pandeia flux units via Synphot for Power Law spectrum. "
                if self.webapp:
                    msg += "(%s)" % type(e)
                else:
                    msg += repr(e)
                raise SynphotError(value=msg)
        if self.unit == 'fnu':
            try:
                sp = syn.spectrum.SourceSpectrum(Empirical1D, points=self.wave << PANDEIA_WAVEUNITS, lookup_table=flux << syn.units.FNU)
                flux = sp(sp.waveset).value
            except Exception as e:
                msg = "Error converting flam units to Pandeia flux units via Synphot for Power Law spectrum. "
                if self.webapp:
                    msg += "(%s)" % type(e)
                else:
                    msg += repr(e)
                raise SynphotError(value=msg)

        return syn.units.convert_flux(self.wave,flux,PANDEIA_FLUXUNITS)


class Flat(Analytic):

    """
    Class implementing a flat continuum in either Fnu or Flam
    """

    def _sanity_checks(self):
        """
        Run sanity checks specify to Flat SEDs
        """
        Analytic._sanity_checks(self)
        for par in ['unit', 'units']:
            if not hasattr(self, par):
                msg = "Missing required parameter %s for %s." % (par, self.__class__.__name__)
                raise DataConfigurationError(value=msg)

        if self.unit not in self.units:
            msg = "Unsupported unit, %s, for %s." % (par, self.__class__.__name__)
            raise EngineInputError(value=msg)

    def create_flux(self):
        if self.unit == 'flam':
            try:
                # Make sure this spectrum is created in FLAM units
                sp = syn.spectrum.SourceSpectrum(ConstFlux1D, amplitude=1*syn.units.FLAM)
                flux = sp(self.wave)
            except Exception as e:
                msg = "Error converting flam units to Pandeia flux units via Synphot for Flat spectrum. "
                if self.webapp:
                    msg += "(%s)" % type(e)
                else:
                    msg += repr(e)
                raise SynphotError(value=msg)
        if self.unit == 'fnu':
            try:
                # Make sure this spectrum is created in FNU units (like mJy)
                sp = syn.spectrum.SourceSpectrum(ConstFlux1D, amplitude=1*PANDEIA_FLUXUNITS)
                flux = sp(self.wave)
            except Exception as e:
                msg = "Error converting flam units to Pandeia flux units via Synphot for Flat spectrum. "
                if self.webapp:
                    msg += "(%s)" % type(e)
                else:
                    msg += repr(e)
                raise SynphotError(value=msg)

        return syn.units.convert_flux(self.wave,flux,PANDEIA_FLUXUNITS)


class BlackBody(Analytic):

    """
    Class implementing a Plank blackbody continuum
    """

    def _sanity_checks(self):
        """
        Run sanity checks specify to blackbody SEDs
        """
        Analytic._sanity_checks(self)
        if not hasattr(self, 'temp'):
            msg = "Missing required parameter 'temp' for %s." % self.__class__.__name__
            raise DataConfigurationError(value=msg)

    def create_flux(self):
        # use astropy analytic function and astropy.units to specify microns and K.
        # flux is in erg cm-2 s-1 hz-1 sr-1, but will be scaled by normalization.
        bb = Black_Body(temperature=self.temp * u.K)
        flux = bb(self.wave)
        # because it's in units that include steradian, we need to multiply by steradian to take them out
        # This will be normalized anyway.
        return flux * u.sr


class No_Continuum(Analytic):

    """
    Class implementing zero continuum (for use with emission lines).
    Implement as just an alias for the base class.
    """

    pass


class Parameterized(SED):

    """
    Class implementing SED's that are calculated from a model using a set of input parameters.
    The parameters can either come from a specfied set contained within a configured catalog
    (usual for webapp mode) or as input configuration data.
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        """
        How an SED is configured depends on the defaults and any input configuration.

        Parameters
        ----------
        webapp: bool
            Determines whether strict API checks should be performed
        config: dict
            Extra configuration information in engine input API dict format
        **kwargs: keyword/value pairs
            Extra configuration information in kwargs format
        """
        SED.__init__(self, webapp=webapp, config=config, **kwargs)
        spectra_file = os.path.join(default_refdata_directory, "sed", self.spectra["config"])
        self.spectra = read_json(spectra_file)
        self.wave, self.flux = self.get_spectrum()
        self.wmin = self.wave.min()
        self.wmax = self.wave.max()

    def _sanity_checks(self):
        """
        Make sure the required parameters are included.
        """
        for p in self.api_parameters:
            if not hasattr(self, p):
                msg = "Missing parameter, %s, for %s" % (p, self.__class__.__name__)
                raise EngineInputError(value=msg)


class Phoenix(Parameterized):

    """
    Class implementing the Phoenix parameterized stellar spectrum models. The inputs to
    the models are:

    teff - effective temperature (K; range 2000 to 70000) 
    metallicity - log10 of the metallicity in units of solar metallicity (range -4.0 to 0.5) 
    log_g - surface gravity (log10 in cgs units; range 0.0 to 5.5)

    If a key is present, it will be used to look up parameters from a configured look-up
    table. Otherwise the model parameters will be taken from default configuration or
    input.
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        self.api_parameters = ["sed_type", "key"]
        self.api_ignore = ["spectrum_id", "teff", "metallicity", "log_g", "comment", "display_string", "spectra", "z"]
        if "teff" in config:
            # if you passed in a configuration with teff/log_g/metallicity in it, we will honor that.
            self.api_parameters = ["sed_type", "teff", "metallicity", "log_g"]
            self.api_ignore = ["spectrum_id", "key",  "comment", "display_string", "spectra", "z"]

        Parameterized.__init__(self, webapp=webapp, config=config, **kwargs)

    def get_spectrum(self):
        """
        Use stsynphot.grid_to_spec() to calculate spectrum and convert to microns/mJy.  If
        self.key is defined, use key to look up a specified set of parameters.  Otherwise,
        get them from the configured attributes.
        """
        if hasattr(self, "teff"):
            m = self.metallicity
            t = self.teff
            g = self.log_g
        else:
            if self.key not in self.spectra:
                msg = "Provided SED key, %s, not supported." % self.key
                raise EngineInputError(value=msg)
            pars = self.spectra[self.key]
            m = pars['metallicity']
            t = pars['teff']
            g = pars['log_g']

        try:
            sp = stsyn.grid_to_spec('phoenix', t, m, g)
        except Exception as e:
            msg = "Error creating Phoenix spectrum via Synphot. "
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise SynphotError(value=msg)
        return sp.waveset, sp(sp.waveset, flux_unit=PANDEIA_FLUXUNITS)


class K93Models(Parameterized):

    """
    Class implementing the Kurucz (1993) parameterized stellar spectrum models. The inputs
    to the models are:

    teff - effective temperature (K; range 3500 to 50000) 
    log_g - surface gravity (log10 in cgs units; range 0.0 to 5.0)

    If used with webapp=True, the parameters will come from a configured look-up table
    using a provided key. Otherwise the model parameters will be taken from default
    configuration or input.
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        self.api_parameters = ["sed_type", "key"]
        self.api_ignore = ["spectrum_id", "teff", "log_g", "comment", "display_string", "spectra", "z"]
        if "teff" in config:
            # if you passed in a configuration with teff/log_g/metallicity in it, we will honor that.
            self.api_parameters = ["sed_type", "teff", "log_g"]
            self.api_ignore = ["spectrum_id", "key",  "comment", "display_string", "spectra", "z"]

        Parameterized.__init__(self, webapp=webapp, config=config, **kwargs)

    def get_spectrum(self):
        """
        Use stsynphot.grid_to_spec() to calculate spectrum and convert to microns/mJy.  If
        self.webapp=True, use key to look up a specified set of parameters.  Otherwise,
        get them from the configured attributes.
        """
        if hasattr(self, "teff") and hasattr(self, "log_g"):
            m = 0
            t = self.teff
            g = self.log_g
        else:
            if self.key not in self.spectra:
                msg = "Provided SED key, %s, not supported." % self.key
                raise EngineInputError(value=msg)
            pars = self.spectra[self.key]
            m = 0
            t = pars['teff']
            g = pars['log_g']
        try:
            sp = stsyn.grid_to_spec('k93models', t, m, g)
        except Exception as e:
            msg = "Error creating K93 model spectrum via Synphot. "
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise SynphotError(value=msg)
        # The k93 output is float32, at least the wavelength needs to be float64 for extinction
        return Quantity(sp.waveset, dtype=np.float64), sp(sp.waveset, flux_unit=PANDEIA_FLUXUNITS)


class CK04Models(Parameterized):

    """
    Class implementing the Castelli & Kurucz (2004) parameterized stellar spectrum models.
    The inputs to the models are:

    teff - effective temperature (K; range 3500 to 50000) 
    log_g - surface gravity (log10 in cgs units; range 0.0 to 5.0)

    If used with webapp=True, the parameters will come from a configured look-up table
    using a provided key. Otherwise the model parameters will be taken from default
    configuration or input.
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        self.api_parameters = ["sed_type", "key"]
        self.api_ignore = ["spectrum_id", "teff", "log_g", "comment", "display_string", "spectra", "z"]
        if "teff" in config:
            # if you passed in a configuration with teff/log_g/metallicity in it, we will honor that.
            self.api_parameters = ["sed_type", "teff", "log_g"]
            self.api_ignore = ["spectrum_id", "key",  "comment", "display_string", "spectra", "z"]

        Parameterized.__init__(self, webapp=webapp, config=config, **kwargs)

    def get_spectrum(self):
        """
        Use stsynphot.grid_to_spec() to calculate spectrum and convert to microns/mJy.  If
        self.webapp=True, use key to look up a specified set of parameters.  Otherwise,
        get them from the configured attributes.
        """
        if hasattr(self, "teff") and hasattr(self, "log_g"):
            m = 0
            t = self.teff
            g = self.log_g
        else:
            if self.key not in self.spectra:
                msg = "Provided SED key, %s, not supported." % self.key
                raise EngineInputError(value=msg)
            pars = self.spectra[self.key]
            m = 0
            t = pars['teff']
            g = pars['log_g']
        try:
            sp = stsyn.grid_to_spec('ck04models', t, m, g)
        except Exception as e:
            msg = "Error creating CK04 model spectrum via Synphot. "
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise SynphotError(value=msg)
        # The ck04 output is float32, at least the wavelength needs to be float64 for extinction
        return Quantity(sp.waveset, dtype=np.float64), sp(sp.waveset, flux_unit=PANDEIA_FLUXUNITS)


class Mapped(SED):

    """
    Class implementing SED's that are looked-up via an input 'key' and loaded from a file.
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        """
        How an SED is configured depends on the defaults and any input configuration.

        Parameters
        ----------
        webapp: bool
            Determines whether strict API checks should be performed
        config: dict
            Extra configuration information in engine input API dict format
        **kwargs: keyword/value pairs
            Extra configuration information in kwargs format
        """
        SED.__init__(self, webapp=webapp, config=config, **kwargs)
        spectra_file = os.path.join(default_refdata_directory, "sed", self.spectra["config"])
        self.spectra = read_json(spectra_file)
        self.wave, self.flux = self.get_spectrum()
        self.wmin = self.wave.min()
        self.wmax = self.wave.max()

    def _sanity_checks(self):
        """
        Make sure a key is provided
        """
        if not hasattr(self, "key"):
            msg = "Must provide a key for looking up a SED."
            raise DataConfigurationError(value=msg)

    def get_spectrum(self):
        """
        Use self.key to grab filename out of self.spectra and load data using Synphot.
        Return wave and flux in microns and mJy.
        """
        if self.key not in self.spectra:
            msg = "Provided SED key, %s, not supported." % self.key
            raise EngineInputError(value=msg)

        filepath = os.path.join(default_refdata_directory,
                                "sed",
                                self.__class__.__name__.lower(),
                                self.spectra[self.key]['filename'])
        try:
            sp = syn.spectrum.SourceSpectrum.from_file(filepath)
        except FileNotFoundError:
            raise DataError("File for SED key {} not found.".format(self.key))
        return sp.waveset, sp(sp.waveset, flux_unit=PANDEIA_FLUXUNITS)


class HST_CalSpec(Mapped):

    """
    Class implementing catalog of HST calibration stars. Purely mapped with no extra functionality.
    """

    pass

class CDBS(Mapped):

    """
    Class implementing SED's that are looked-up from CDBS via an input 'key' and loaded from a file.
    """

    def get_spectrum(self):
        """
        Use self.key to grab filename out of self.spectra and load data using Synphot.
        Return wave and flux in microns and mJy.
        """
        if self.key not in self.spectra:
            msg = "Provided SED key, %s, not supported." % self.key
            raise EngineInputError(value=msg)

        filepath = os.path.join(os.environ['PYSYN_CDBS'],
                                self.spectra[self.key]['filename'])
        try:
            if filepath.endswith("fits") or filepath.endswith("fit"):
                sp = syn.spectrum.SourceSpectrum.from_file(filepath)
            else:
                sp = syn.spectrum.SourceSpectrum.from_file(filepath, guess=False)
        except FileNotFoundError:
            raise DataError("File for SED key {} not found.".format(self.key))
        return sp.waveset, sp(sp.waveset, flux_unit=PANDEIA_FLUXUNITS)

class Custom(CDBS):

    """
    Class implementing custom user-specified CDBS spectra
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        """
        How an SED is configured depends on the defaults and any input configuration.

        Parameters
        ----------
        webapp: bool
            Determines whether strict API checks should be performed
        config: dict
            Extra configuration information in engine input API dict format
        **kwargs: keyword/value pairs
            Extra configuration information in kwargs format
        """
        self.api_parameters = ["sed_type", "key"]
        SED.__init__(self, webapp=webapp, config=config, **kwargs)
        self.spectra = {self.key: {"filename": self.key}} # the key IS the filename
        self.wave, self.flux = self.get_spectrum()
        self.wave = Quantity(self.wave, dtype=np.float64)
        self.wmin = self.wave.min()
        self.wmax = self.wave.max()

class Brown2014(CDBS):

    """
    Class implementing catalog of integrated galaxy spectra from Brown et al. (2014).
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        """
        How an SED is configured depends on the defaults and any input configuration.

        Parameters
        ----------
        webapp: bool
            Determines whether strict API checks should be performed
        config: dict
            Extra configuration information in engine input API dict format
        **kwargs: keyword/value pairs
            Extra configuration information in kwargs format
        """
        SED.__init__(self, webapp=webapp, config=config, **kwargs)
        spectra_file = os.path.join(default_refdata_directory, "sed", self.spectra["config_all"])
        self.spectra = read_json(spectra_file)
        self.wave, self.flux = self.get_spectrum()
        self.wmin = self.wave.min()
        self.wmax = self.wave.max()

class Novae(CDBS):

    """
    Novae require nothing beyond the standard CDBS class.
    """

    pass

class PNE(CDBS):

    """
    Planetary Nebulae require nothing beyond the standard CDBS class.
    """

    pass

class Sun_Planets(CDBS):

    """
    Solar System objects require nothing beyond the standard CDBS class.
    """

    pass

class Cool_Dwarfs(CDBS):

    """
    Cool Dwarf spectra require nothing beyond the standard CDBS class.
    """

    pass

class Brown2019(CDBS):

    """
    Brown et al. (2019) galaxy spectra require nothing beyond the standard CDBS class.
    """

    pass

class SWIRE(CDBS):

    """
    Ordinary galaxy spectra require nothing beyond the standard CDBS class.
    """

    pass

class Comp_QSO(CDBS):

    """
    Composite QSO spectra require nothing beyond the standard CDBS class.
    """

    pass

class Stellar_Pop(CDBS):

    """
    Stellar Population spectra require nothing beyond the standard CDBS class.
    """

    pass

class Pickles(CDBS):

    """
    The Pickles Stellar Spectral Flux Library (Pickles, A.J. 1998) requires nothing beyond
    the standard CDBS class.
    """

    pass

class BZ77(CDBS):

    """
    The Bruzual Atlas requires nothing beyond the standard CDBS class.
    """

    pass

class NonStellar(CDBS):

    """
    The non-stellar spectra class requires nothing beyond the standard CDBS class.
    """

    pass

class BT_Settl(CDBS):

    """
    The BT Settl models requires nothing beyond the standard CDBS class.
    """

    pass


def SEDFactory(webapp=False, config={}, **kwargs):
    """
    Function to take configuration data and build/return an appropriately configured instance
    of the desired SED class.

    Parameters
    ----------
    webapp: bool
        Switch to toggle strict API checking
    config: dict
        Configuration data in engine API dict format
    **kwargs: keyword/value pairs
        Additional configuration data
    """
    all_config = merge_data(config, dict(**kwargs))
    types = recursive_subclasses(SED)
    sed_types = [t.__name__.lower() for t in types]
    type_map = dict(list(zip(sed_types, types)))

    # pop the sed_format out of the input config and error out if it's not there
    try:
        sed_type = all_config["sed_type"]
    except KeyError as e:
        msg = "No SED format specified. (%s)" % e
        raise EngineInputError(value=msg)

    # make sure sed_format is a supported one and build/return appropriate instance if so
    if sed_type not in sed_types:
        msg = "Unsupported or not yet implemented SED: %s" % sed_type
        raise EngineInputError(value=msg)
    else:
        cls = type_map[sed_type](webapp=webapp, config=all_config)
        return cls



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import bisect
import copy
import functools
import os

import astropy.io.fits as fits
import numpy as np

from .custom_exceptions import DataError

class PSFLibrary:

    """
    Class for encapsulating a hard PSF library, typically one generated by WebbPSF or its derivatives.

    Parameters
    ----------
    instrument : Instrument
        An instrument-class object that can hold instrument-specific information for PSF selection.
    path : string, optional
        the path to the PSF library root (any .fits files in the
        path root will be assumed to be a valid PSF file in the WebbPSF format.)
        If it is none, the PSF library from the package distribution will be used.
    psf_key : string, optional
        restrict psfs to a specific mode with the name matching the keyword string.
        Strictly, the reader will look for psf file names with a substring matching the aperture
        string. 'all' reads all psfs in the path. This is the default. The aperture string in the filename
        should match the .fits keyword APERTURE.

    Methods
    -------
    read_library
    get_values
    get_good_psfs
    get_psf

    """

    def __init__(self, wavelength_range, path=None, psf_key='all', x_offset=0):
        self.wavelength_range = wavelength_range
        if path is None:
            path = os.path.join(os.path.dirname(__file__), 'refdata', 'psfs')
        if 'masklwb' in psf_key:
            self.x_offset = -1*x_offset
        else:
            self.x_offset = x_offset

        self.read_library(path, psf_key)
        self._psfs = sorted(self._psfs, key=lambda psf: psf["wave"])

    def read_library(self, path, aperture, wave_unit='m'):
        """
        Read in a library of PSFs stored in FITS files. The filenames contain information about each PSF. The fields are
        separated by _'s with the resulting list entries corresponding to:
            0 - instrument name
            1 - apertures for which PSF is used. this is a - separated string that is split() into a list of apertures
            2 - wavelength at which PSF is calculated
            3, 4 (optional) - x,y rectangular coordinates in FOV at which at which PSF is calculated

        Parameters
        ----------
        path: string
            Path containing the PSF FITS images

        aperture: string
            The key that selects the appropriate PSFs for this calculation

        wave_unit: string
            Wavelength unit used by PSF library (default: m)
        """
        files = os.listdir(path)
        files = [x for x in files if x.endswith(".fits") ] # Filter to just .fits files

        psf_files = []
        filename_waves = []
        for file in files:
            filename, ext = os.path.splitext(file)
            filename_split = filename.split('_')
            filename_instrument = filename_split[0]
            filename_apertures = filename_split[1].split('-')
            filename_wave = float(filename_split[2])
            # If the psf is on-axis, the offset coordinates may not exist

            if aperture == 'all':
                psf_files.append(file)
                filename_waves.append(filename_wave)
            elif aperture in filename_apertures:
                if 'masklwb' in aperture or 'maskswb' in aperture:
                    # for NIRCam bar PSFs, they not only DO exist, we should ignore all
                    # PSFs far from the bar_offset
                    filename_xoff = float(filename_split[3])
                    # Only load on-grid PSFs within +/- 2.5 arcsec of the optimal position, and this particular filter's optimal position. Skip the rest.
                    if np.abs(filename_xoff - self.x_offset) > 2.5 or ((filename_xoff % 1 != 0) and (np.abs(filename_xoff - self.x_offset) > 0.01)):
                        continue
                psf_files.append(file)
                filename_waves.append(filename_wave)
            # If the aperture has a string with an underscore in it, try to use only the first part.
            elif aperture.split('_')[0] in filename_apertures:
                psf_files.append(file)
                filename_waves.append(filename_wave)
            else:
                pass

        self._psfs = []

        # Only load the PSFs we need for this calculation. That means finding the PSFs just above and
        # below the wavelength range
        filename_waves = sorted(list(set(filename_waves)))
        min_wave_idx = max(bisect.bisect(filename_waves, self.wavelength_range['wmin'])-1, 0)
        max_wave_idx = min(bisect.bisect(filename_waves, self.wavelength_range['wmax']), len(filename_waves)-1)

        min_wave = filename_waves[min_wave_idx]
        max_wave = filename_waves[max_wave_idx]

        for psf_file in psf_files:
            # only load if it's within the wavelength range we care about
            filename, ext = os.path.splitext(psf_file)
            file_wavelength = float(filename.split('_')[2])
            if file_wavelength >= min_wave and file_wavelength <= max_wave:

                hdulist = fits.open(path + '/' + psf_file, memmap=False)

                psf_int = copy.deepcopy(hdulist[0].data)

                ins = hdulist[0].header['INSTRUME'].lower()
                nwaves = hdulist[0].header['NWAVES']
                if nwaves != 1:
                    raise ValueError("Input PSF must be monochromatic (nwaves = #r)" % nwaves)

                wave_scl = self._get_unit_scl(wave_unit)
                wave = hdulist[0].header['WAVE0'] * wave_scl
                    
                if "PIXSCLX" in hdulist[0].header:
                    pix_sclx = hdulist[0].header['PIXSCLX']
                    pix_scly = hdulist[0].header['PIXSCLY']
                elif "PIXELSCL" in hdulist[0].header:
                    pix_sclx = hdulist[0].header['PIXELSCL']
                    pix_scly = hdulist[0].header['PIXELSCL']
                else:
                    raise DataError("Missing pixel scale keyword")
                diff_limit = hdulist[0].header['DIFFLMT']
                aperture_name = hdulist[0].header['APERTURE'].lower()
                pup_thru = hdulist[0].header['PUP_THRU']
                psfsum = np.sum(psf_int)
                # offset_x/offset_y is the offset from the optical center in webbpsf.  However, for NIRCam masklwb
                # and maskswb, this includes the offset from the optical center to the optimal position along the bar.
                # what we need for source association is instead the offset from that optimal position. this is stored in the
                # optoff_x/optoff_y keywords which need to be provided in these cases. aperture_name here will have these
                # aperture and filter names concatenated, so do a string compare to look for aperture.
                try:
                    offset_x = hdulist[0].header['OPTOFF_X']
                    offset_y = hdulist[0].header['OPTOFF_Y']
                except KeyError:
                    offset_x = 0.0
                    offset_y = 0.0

                upsamp = hdulist[0].header['DET_SAMP']

                psf = {
                    'int': psf_int,
                    'wave': wave,
                    'pix_sclx': pix_sclx,
                    'pix_scly': pix_scly,
                    'diff_limit': diff_limit,
                    'upsamp': upsamp,
                    'instrument': ins,
                    'aperture_name': aperture_name,
                    'source_offset': (offset_x, offset_y),
                    'pupil_throughput': pup_thru,
                    'psfsum': psfsum
                }

                self._psfs.append(psf)
                hdulist.close()

    @functools.lru_cache(maxsize=None)
    def get_values(self, key, instrument, aperture_name, source_offset=None):
        """
        Returns the available values of a given key for a given instrument and aperture name.

        Parameters
        ----------
        key: string
            Desired key. One of 'int', 'wave', 'pix_sclx', 'pix_scly', 'diff_limit', 'upsamp', 
            'instrument', 'aperture_name', 'source_offset_x', or 'source_offset_y'

        Returns
        -------
        result: tuple of format (list, list)
            Lists of index id's and values
        """
        if source_offset is None:
            # because we cannot reference self within the method signature, we have to set a default here.
            source_offset = (int(np.round(self.x_offset)),0)
        # any part of the aperture name after a double underscore should be ignored when dealing with the PSF library
        aperture_name = aperture_name.split("__")[0]
        values = [psf[key] for psf in self._psfs
                  if (instrument == psf['instrument'] and
                      aperture_name in psf['aperture_name']) and
                      source_offset == psf['source_offset']]
        ids = [i for i, psf in enumerate(self._psfs)
               if (instrument == psf['instrument'] and
                   aperture_name in psf['aperture_name']) and
                   source_offset == psf['source_offset']]
        result = ids, values
        return result

    def get_psf(self, wave, instrument, aperture_name, source_offset=(0, 0)):
        """
        Get PSF given wavelength, instrument, instrument aperture, and optionally source xy offset coordinates.
        Values are interpolated linearly between the two PSF library entries that bracket the given wavelength.

        Parameters
        ----------
        wave: float
            Desired wavelength for the PSF
        instrument: str
            Instrument name
        aperture_name: str
            Name of the instrument aperture

        Returns
        -------
        psf: dict
            Dict containing the PSF and associated information
        """
        wids, psf_waves = self.get_values('wave', instrument, aperture_name, source_offset=source_offset)
        nids, nearest_waves = self._find_two_nearest(psf_waves, wave)

        # if wave is not in psf_waves then the two closest will be returned
        if len(nids) == 2:
            ids = [wids[nids[0]], wids[nids[1]]]

            pix_sclx0 = self._psfs[ids[0]]['pix_sclx']
            pix_sclx1 = self._psfs[ids[1]]['pix_sclx']
            pix_scly0 = self._psfs[ids[0]]['pix_scly']
            pix_scly1 = self._psfs[ids[1]]['pix_scly']
            if pix_sclx0 != pix_sclx1:
                raise ValueError("Pixel scales in the library must be the same for a single instrument aperture.")
            if pix_scly0 != pix_scly1:
                raise ValueError("Pixel scales in the library must be the same for a single instrument aperture.")

            psf_int0 = self._psfs[ids[0]]['int']
            psf_int1 = self._psfs[ids[1]]['int']
            wave0 = self._psfs[ids[0]]['wave']
            wave1 = self._psfs[ids[1]]['wave']
            diff_limit0 = self._psfs[ids[0]]['diff_limit']
            diff_limit1 = self._psfs[ids[1]]['diff_limit']

            upsamp0 = self._psfs[ids[0]]['upsamp']
            upsamp1 = self._psfs[ids[1]]['upsamp']
            if upsamp0 != upsamp1:
                raise ValueError("Upsampling factors in the library must be the same for a single instrument aperture.")

            psf_int = psf_int0 + (psf_int1 - psf_int0) * (wave - wave0) / (wave1 - wave0)
            diff_limit = diff_limit0 + (diff_limit1 - diff_limit0) * (wave - wave0) / (wave1 - wave0)

        # if wave is one of the values in psf_waves then only it will be returned and
        # no interpolation is necessary
        else:
            ids = nids

            psf_int = self._psfs[ids[0]]['int']
            diff_limit =  self._psfs[ids[0]]['diff_limit']
            pix_sclx0 = self._psfs[ids[0]]['pix_sclx']
            pix_scly0 = self._psfs[ids[0]]['pix_scly']

            upsamp0 = self._psfs[ids[0]]['upsamp']

        psf = {
            'int': psf_int,
            'wave': wave,
            'pix_sclx': pix_sclx0,
            'pix_scly': pix_scly0,
            'diff_limit': diff_limit,
            'upsamp': upsamp0,
            'instrument': self._psfs[ids[0]]['instrument'],
            'aperture_name': self._psfs[ids[0]]['aperture_name'],
            'source_offset': self._psfs[ids[0]]['source_offset'],
            'pupil_throughput': self._psfs[ids[0]]['pupil_throughput']
        }
        return psf

    def get_pix_scale(self, instrument, aperture_name):
        """
        Get PSF pixel scale for given instrument/aperture

        Parameters
        ----------
        instrument: str
            Instrument name
        aperture_name: str
            Name of instrument aperture

        Returns
        -------
        pix_scl: float
            Pixel scale of the PSF in arcsec/pixel
        """
        wids, psf_waves = self.get_values('wave', instrument, aperture_name)
        pix_sclx = self._psfs[wids[0]]['pix_sclx']
        pix_scly = self._psfs[wids[0]]['pix_scly']
        return pix_sclx, pix_scly

    def get_det_scale(self, instrument, aperture_name):
        """
        Get detector pixel scale for given instrument/aperture, which is 
        the pixel scale without oversampling (or, after binning)

        Parameters
        ----------
        instrument: str
            Instrument name
        aperture_name: str
            Name of instrument aperture

        Returns
        -------
        pix_scl: float
            Pixel scale of the detector in arcsec/pixel
        """
        wids, psf_waves = self.get_values('wave', instrument, aperture_name)
        pix_sclx = self._psfs[wids[0]]['pix_sclx'] * self._psfs[wids[0]]['upsamp']
        pix_scly = self._psfs[wids[0]]['pix_scly'] * self._psfs[wids[0]]['upsamp']
        return pix_sclx, pix_scly

    def get_shape(self, instrument, aperture_name):
        """
        Get PSF shape for given instrument/aperture

        Parameters
        ----------
        instrument: str
            Instrument name
        aperture_name: str
            Name of instrument aperture

        Returns
        -------
        sh: tuple (int, int)
            Shape of the PSF kernel image
        """
        try:
            wids, psf_waves = self.get_values('wave', instrument, aperture_name)
            sh = self._psfs[wids[0]]['int'].shape
        except IndexError:
            raise DataError("No valid PSFs found")
        return sh

    def get_upsamp(self, instrument, aperture_name):
        """
        Get PSF upsampling for given instrument/aperture

        Parameters
        ----------
        instrument: str
            Instrument name
        aperture_name: str
            Name of instrument aperture

        Returns
        -------
        upsamp: int
            PSF upsampling factor
        """
        wids, psf_waves = self.get_values('wave', instrument, aperture_name)
        upsamp = self._psfs[wids[0]]['upsamp']
        return upsamp

    def get_offsets(self, instrument, aperture_name):
        """
        Get available PSF offset positions. PSF libraries supporting position-dependent
        PSFs will have multiple. Libraries that do not support position-dependent PSFs will
        only have one position.

        Parameters
        ----------
        list: list-like
            List of values to search
        value: float
            Input value to be bracketed

        Returns
        -------
        psf_positions: list of tuples (source_offset_x, source_offset_y)
            Polar coordinates of PSF offsets.

        """
        aperture_name = aperture_name.split("__")[0]
        offsets = [psf['source_offset'] for psf in self._psfs if instrument ==
                   psf['instrument'] and aperture_name in psf['aperture_name']]
        unique_offsets = list(set(offsets))
        return unique_offsets

    def get_pupil_throughput(self, wave, instrument, aperture_name):
        """
        Get the pupil throughput for a given instrument and aperture. Throughputs are
        calculated for the shortest wavelength in the set, and are given in the headers
        of every fits file.

        Parameters
        ----------
        wave: float
            Wavelength of interest
        instrument: string
            Name of the instrument
        aperture_name: string
            String used to select the PSFs. Nominally, this is the aperture, but in certain cases it's not.

        Returns
        -------
        pupil_thru: float
            Pupil throughput fraction
        """
        # get pupil losses from the PSF pupil_throughput keyword
        psf_offsets = self.get_offsets(instrument, aperture_name)
        unique_positions = np.unique([psf_offset[0] for psf_offset in psf_offsets])
        # if there's only one PSF offset per wavelength, just grab the one associated with the first wavelength (they're
        # all the same)
        if unique_positions.size == 1:
            pupil_thru = self.get_psf(wave,instrument,aperture_name)['pupil_throughput']
        # if there are multiple PSF offsets, find the one for the unocculted PSF.
        else:
            offsets_x = np.asarray(list(zip(*psf_offsets))[0])
            offsets_y = np.asarray(list(zip(*psf_offsets))[1])
            # we need to select the maximum y for the maximum x; for NIRCam MASKLWB the maximum x is at the smallest
            # part of the wedge.
            x_off = np.asarray([k for k in range(len(offsets_x)) if offsets_x[k] == np.max(offsets_x)])
            psf_offset_x = np.max(offsets_x)
            psf_offset_y = np.max(offsets_y[x_off])
            pupil_thru = self.get_psf(wave, instrument, aperture_name, source_offset=(psf_offset_x,psf_offset_y))['pupil_throughput']
        return pupil_thru

    def _calculate_distances(self, psf_offsets, source):
        """
        Compute the distances between a source and the possible PSF offsets.

        Parameters
        ----------
        psf_offsets: np.array
            The possible PSF offsets for this instrument and aperture combination.
        source: list
            The x and y coordinates of the source to be matched.

        Returns
        -------
        distances: np.array
            The distances between the source and the psf_offsets
        """
        distances = [np.sqrt((source[0] - psf_offset[0]) ** 2 + \
                             (source[1] - psf_offset[1]) ** 2) \
                     for psf_offset in psf_offsets]
        return distances

    def associate_offset_to_source(self, sources, instrument, aperture_name):
        """
        Pandeia does not interpolate spatially between PSFs; it merely picks the closest offset to the source.
        This is the function that makes that selection.

        Parameters
        ----------
        sources: list
            A list of Source objects
        instrument: string
            The name of the instrument
        aperture_name:
            String used to select the PSFs. Nominally this is the aperture, but in certain cases, it is changed in the
            _loadpsfs functions in jwst.py, hst.py, and roman.py.

        Returns
        -------
        psf_associations: list
            The X and Y coordinates of the associated PSF.
        """

        psf_offsets = self.get_offsets(instrument, aperture_name)

        # Make sure there are offsets given the instrument and aperture
        if len(psf_offsets) == 0:
            raise DataError("No psf offsets found for instrument {} and aperture {}".format(instrument, aperture_name))

        unique_positions = np.unique([psf_offset[0] for psf_offset in psf_offsets])

        psf_associations = []
        for source in sources:
            if unique_positions.size == 1:
                # if there is only one x position, we will assume azimuthal symmetry, and only use the y positions
                source_radius = np.sqrt(source.position['x_offset'] ** 2 + source.position['y_offset'] ** 2)
                distances = [np.abs(psf_offset[1] - source_radius) for psf_offset in psf_offsets]
            else:
                # multiple x positions indicate that we'd expect the full 2D field
                # to be appropriately sampled by the PSF library.
                if 'masklwb' in aperture_name or 'maskswb' in aperture_name:
                    offset_y = np.abs(source.position['y_offset'])
                    offset_x = source.position['x_offset'] + self.x_offset
                if 'fqpm' in aperture_name:
                    # first, the FQPM masks (and their PSF library) are rotated by 5 degrees cw relative to the
                    #  detector axes. Therefore, we need to rotate 5 degrees ccw to get the real positions relative to
                    #  THAT grid.
                    off_x = source.position['x_offset'] * np.cos(5.0 * np.pi / 180.0) - source.position['y_offset'] * np.sin(5.0 * np.pi / 180.0)
                    off_y = source.position['x_offset'] * np.sin(5.0 * np.pi / 180.0) + source.position['y_offset'] * np.cos(5.0 * np.pi / 180.0)
                    # second, we take the absolute value of both rotated coordinates, AND reflect across the x=y axis
                    if np.abs(off_y) < np.abs(off_x):
                        offset_x = np.abs(off_x)
                        offset_y = np.abs(off_y)
                    else:
                        offset_x = np.abs(off_y)
                        offset_y = np.abs(off_x)

                distances = self._calculate_distances(psf_offsets, [offset_x,offset_y])

            closest_index = np.argmin(distances)
            psf_associations.append(psf_offsets[closest_index])

        return psf_associations

    def _find_two_nearest(self, values, value):
        """
        Find the subscripts of the two neighboring values in list relative to
        an input value.

        Parameters
        ----------
        values: list-like
            List of values to search
        value: float
            Input value to be bracketed

        Returns
        -------
        vals: tuple (list, np.ndarray)
            Bracketing indices in both list and numpy array format
        """
        nplist = np.array(values)

        # ensure that value is not outside the range of list.
        # rounding value of 6 exceeds the highest number of decimal places that actually 
        # exists in the data directories (5, as of 2022-03-21)
        if np.round(value,6) < np.round(nplist.min(),6) or value > np.round(nplist.max(),6):
            raise DataError("Specified value {} must be within supported range ({}-{})".format(value, np.round(nplist.min(),6),np.round(nplist.max(),6)))

        # This function is called from get_psf(), which can handle getting only one PSF 
        # back in the case that we don't need interpolation because an actual PSF exists 
        # at that wavelength.
        if np.round(nplist.min(),6) == np.round(value,6):
            vals = [nplist.argmin()],[nplist.min()]
        elif np.round(nplist.max(),6) == np.round(value,6):
            vals = [nplist.argmax()], [nplist.max()]
        # This code will not work if we're requesting exactly the highest wavelength or 
        # lowest wavelength that has actually been generated.
        else:
            diff = nplist - value
            diff_sorted = np.sort(diff)

            smaller = np.max(diff_sorted[diff_sorted<=0])
            larger = np.min(diff_sorted[diff_sorted>=0])

            ids = [i for i, v in enumerate(diff) if diff[i] == smaller or diff[i] == larger]

            vals = ids, nplist[ids]

        return vals

    def _get_unit_scl(self, unit):
        """
        Get unit scale factor for given unit string: m, micron, nm, or Angstrom.
        This should get replaced with real astropy.units support eventually.

        Parameters
        ----------
        unit: str
            Unit string

        Returns
        -------
        scale: float
            Scale factor corresponding to unit string
        """
        scales = {'m': 1.0e6, 'micron': 1.0, 'nm': 1.0e-3, 'Angstrom': 1.0e-4}
        try:
            scale = scales[unit]
            return scale
        except KeyError:
            raise KeyError('Unknown wavelength unit')
